#include <string.h>
#include <ctype.h>
#include "bsp.h"
#include "tty.h"
#include "flash.h"
#include "power.h"
#include "led.h"
#include "uip/uip.h"

static byte init_mac_addr[] = { 0x00, 0x18, 0xf3, 0x9d, 0x84, 0x20 };
static byte init_ip_addr[] = { 207,22,48,181 };
static byte init_gw_addr[] = { 207,22,48,161 };
static byte init_netmask[] = { 255,255,255,224 };

#define flash_temp (*(DataFlash*)&uip_buf)
static DataFlash *flash_ptr;
#define flash_bytes ((char *)flash_ptr)

typedef struct VarData {
  int offset;
  int size;
  char *name;
  const struct VarData *sub;
} VarData;

#define offsetof(st, m) \
    ((int) ( (char *)&((st *)(0))->m - (char *)0 ))

#define C(a,b) a##b
#define VDE(field)     { offsetof(VStruct, field), sizeof((*(VStruct*)0).field), \
      #field, 0 }
#define VDA(field,idx) { offsetof(VStruct, field[idx]), sizeof((*(VStruct*)0).field[idx]), \
      #field #idx, C(field,_vars) }
#define VDS(field)     { offsetof(VStruct, field), sizeof((*(VStruct*)0).field), \
      #field, C(field,_vars) }
#define VDX { 0, 0, 0, 0 }


#define VStruct DF_Header
static const VarData header_vars[] = {
  VDE(magic),
  VDE(id),
  VDE(poll_sec),
  VDX
};
#undef VStruct

#define VStruct DF_Ether
static const VarData ether_vars[] = {
  VDE(mac_addr),
  VDE(ip_addr),
  VDE(gw_addr),
  VDE(netmask),
  VDX
};
#undef VStruct

#define VStruct DF_Power
static const VarData power_vars[] = {
  VDE(apos),
  VDE(ch1os),
  VDE(ch2os),
  VDE(gain),
  VDE(irmsos),
  VDE(linecyc),
  VDE(mode),
  VDE(phcal),
  VDE(vadiv),
  VDE(vagain),
  VDE(vrmsos),
  VDE(wdiv),
  VDE(wgain),
  VDE(scale),
  VDX
};
#undef VStruct

#define VStruct DataFlash
static const VarData flash_vars[] = {
  VDS(header),
  VDS(ether),
  VDA(power,0),
  VDA(power,1),
  VDA(power,2),
  VDA(power,3),
  VDA(power,4),
  VDA(power,5),
  VDA(power,6),
  VDA(power,7),
  VDA(power,8),
  VDA(power,9),
  VDA(power,10),
  VDA(power,11),
  VDA(power,12),
  VDA(power,13),
  VDA(power,14),
  VDA(power,15),
  VDX
};
#undef VStruct

static const char check_flash_buffer_size[(UIP_BUFSIZE >= DF_SIZE) ? 1 : -1];
static const char check_flash_data_size[(sizeof(DataFlash) == DF_SIZE) ? 1 : -1];

static void
show_var (int ofs, const VarData *vd, int indent);

static void
show_struct (int ofs, const VarData *vd, int indent)
{
  char *data = (char *)flash_ptr + ofs;
  int i, d;
  for (i=0; vd[i].name; i++)
    show_var (ofs, & vd[i], indent);
}

static void
show_var (int ofs, const VarData *vd, int indent)
{
  char *data;
  int d;

  ofs += vd->offset;
  data = (char *)flash_ptr + ofs;
  for (d=0; d<indent; d++)
    tty_putc(' ');
  cprintf("%w", ofs);
  if (!vd->sub)
    {
      tty_puts(" :");
      for (d=0; d<vd->size; d++)
	cprintf(" %b", data[d]);
      for (d=0; d<(6-vd->size)*3; d++)
	tty_putc(' ');
    }
  cprintf(" : %s\n",  vd->name);
  if (vd->sub)
    show_struct (ofs, vd->sub, indent+4);
}

static void
about_to_modify ()
{
  if (flash_ptr == &flash_temp)
    return;
  flash_ptr = &flash_temp;

  memcpy (&flash_temp, &flash_data, sizeof(DataFlash));
}

void
flash_data_init ()
{
  int i;
  about_to_modify ();
  memset (flash_ptr, 0, sizeof(DataFlash));
  flash_ptr->header.id = 0x44332211L;
  flash_ptr->header.poll_sec = 1;
  memcpy (&flash_ptr->ether.mac_addr, &init_mac_addr, 6);
  memcpy (&flash_ptr->ether.ip_addr, &init_ip_addr, 4);
  memcpy (&flash_ptr->ether.gw_addr, &init_gw_addr, 4);
  memcpy (&flash_ptr->ether.netmask, &init_netmask, 4);
  for (i=0; i<16; i++)
    {
      flash_ptr->power[i].gain = GAIN_V1 | GAIN_S05 | GAIN_I1;
      flash_ptr->power[i].mode = DISCF | DISSAG | CYCMODE;
      flash_ptr->power[i].linecyc = 120;
      flash_ptr->power[i].phcal = 0x0d;
      flash_ptr->power[i].scale = 1;
    }
}

void
flash_init ()
{
  int i;
  flash_ptr = &flash_data;
#if 0
  /* This is called from powermeter.c */
  if (flash_data.header.magic != FLASH_MAGIC)
    {
      cprintf("\033[31mFlash data uninitialized!  Initializing... (don't forget to save it)\033[0m\n");
      flash_data_init();
      flash_cmdline();
    }
#endif
}

static void
get_line (char *cmdline, int len)
{
  int idx = 0;

  while (1)
    {
      int ch;
      long timeout = tics_long;

      while (!tty_ready())
	/* 60 sec timeout should be enough ;-) */
	if ((tics_long - timeout) > 60000L)
	  {
	    strcpy (cmdline, "exit");
	    tty_puts(" <timeout> \n");
	    return;
	  }

      ch = tty_readc ();
      switch (ch)
	{
	case 3:
	  strcpy (cmdline, "exit");
	  tty_puts(" Ctrl-C \n");
	  return;
	case 10:
	case 13:
	  cmdline[idx] = 0;
	  tty_putc('\n');
	  return;
	case 8:
	case 127:
	  if (idx > 0)
	    {
	      tty_puts ("\010 \010");
	      idx --;
	    }
	  else
	    tty_putc(7);
	  break;
	case 9:
	  /* ignore tab for now */
	  break;
	default:
	  if (idx < len-1 && isprint(ch))
	    {
	      cmdline[idx] = ch;
	      tty_putc (ch);
	      idx++;
	    }
	  else
	    tty_putc(7);
	}
    }
}

static const VarData *
str2var (char *str, int *ofs)
{
  char *dot;
  char *oldstr = "top";
  int i;
  const VarData *vd = flash_vars;

  *ofs = 0;

  while (1)
    {
      dot = strchr (str, '.');
      if (dot)
	*dot = 0;

      for (i=0; vd[i].name; i++)
	if (strcmp (str, vd[i].name) == 0)
	  break;

      if (vd[i].name == 0)
	{
	  cprintf("name %s not found in %s\n", str, oldstr);
	  return 0;
	}

      if (dot)
	{
	  if (vd[i].sub == 0)
	    {
	      cprintf("%s doesn't have sub-vars\n", str);
	      return 0;
	    }
	  *dot = '.';
	  oldstr = str;
	  str = dot+1;
	  *ofs += vd[i].offset;
	  vd = vd[i].sub;
	}
      else
	return vd+i;
    }
}

static int
do_m (int start, const char *str, int max)
{
  int byte;
  const char *ends;
  while (1)
    {
      byte = strtol (str, &ends, 0);
      if (ends == str)
	return start;
      flash_bytes[start++] = byte;
      if (--max == 0)
	return start;
      str = ends;
      if (*str)
	str++;
    }
}

static void
save_flash_data ()
{
  int i;
  int twiddle = 0;

  cprintf("erasing flash... ");

  /* Enable ER1 rewrite mode.  */
  asm("fclr i");
  fmr0.fmr01 = 0;
  fmr0.fmr01 = 1;
  fmr1.fmr11 = 0;
  fmr1.fmr11 = 1;

  /* Erase block A */
#define FLASH_ERASE_BYTE (*(volatile byte *)(DFLASH1_ADDR + 1024 - 1))
  asm("# pre-erase");
  FLASH_ERASE_BYTE = 0x50;
  FLASH_ERASE_BYTE = 0x20;
  FLASH_ERASE_BYTE = 0xd0;
  asm("# post-erase");

  asm("fset i");

  if (fmr0.fmr07)
    {
      FLASH_ERASE_BYTE = 0x50; /* clear status register */
      fmr0.b = 0x01;	/* disable rewrite mode */
      cprintf("error\n");
      return;
    }
  cprintf("complete\nwriting data...  ");
  fmr0.b = 0;	/* disable rewrite mode */

  /* Enable ER1 rewrite mode.  */
  asm("fclr i");
  fmr0.fmr01 = 0;
  fmr0.fmr01 = 1;
  fmr1.fmr11 = 0;
  fmr1.fmr11 = 1;

#define FLASH ((volatile char *)DFLASH1_ADDR)
  FLASH[i] = 0x50;
  for (i=0; i<sizeof(DataFlash); i++)
    {
      register byte v = ((byte *)&flash_temp)[i];
      volatile byte *fp = ((volatile byte *)DFLASH1_ADDR) + i;
      if (v == 0xff)
	continue;
      fmr4.fmr40 = 1;
      asm("# pre-write");
      *fp = 0x40;
      *fp = v;
      asm("# post-write");

      if (fmr0.fmr06)
	{
	  FLASH[i] = 0x50;
	  fmr0.b = 0x01;	/* disable rewrite mode */
	  asm("fset i");
	  cprintf("error\n");
	  return;
	}
    }

  asm("fset i");
  cprintf("done.\n");
  fmr0.b = 0;	/* disable rewrite mode */
}

void
flash_cmdline ()
{
  char cmdline[80];
  char *tokens[5];
  int idx, nt;

  cprintf("Flash data: %d bytes\n", sizeof(DataFlash));

  while (1)
    {
      tty_puts("\n\033[32mFlash: \033[36m");

      get_line (cmdline, sizeof(cmdline));
      tty_puts("\033[0m");

      /*cprintf("cmdline `%s'\n", cmdline);*/

      {
	int in_word = 0;
	nt = 0;

	for (idx=0; cmdline[idx] && nt < 5; idx++)
	  if (in_word)
	    {
	      if (!isgraph(cmdline[idx]))
		{
		  cmdline[idx] = 0;
		  in_word = 0;
		}		
	    }
	  else
	    {
	      if (isgraph(cmdline[idx]))
		{
		  tokens[nt++] = cmdline+idx;
		  in_word = 1;
		}
	    }

	/*for (idx=0; idx<nt; idx++)
	  cprintf("token [%d] = `%s'\n", idx, tokens[idx]);*/
      }

      if (nt == 0)
	continue;

      if (strcmp (tokens[0], "exit") == 0)
	break;

      else if (strcmp (tokens[0], "diag") == 0)
	{
	  int c;
	  if (nt < 2)
	    {
	      ether_diag (1);
	      for (c=0; c<16; c++)
		power_diag(c, 1);
	    }
	  else if (strcmp (tokens[1], "eth") == 0)
	    ether_diag (0);
	  else
	    {
	      c = atoi (tokens[1]);
	      power_diag (c, 0);
	    }
	}

      else if (strcmp (tokens[0], "show") == 0)
	{
	  if (nt > 1)
	    {
	      int ofs;
	      const VarData *vd = str2var(tokens[1], &ofs);
	      if (!vd)
		continue;
	      show_var (ofs, vd, 4);
	    }
	  else
	    show_struct (0, flash_vars, 4);
	}

      else if (strcmp (tokens[0], "d") == 0)
	{
	  int i, r, o;
	  int start = 0, len = sizeof(DataFlash);
	  int sr, er;

	  if (nt > 1)
	    {
	      start = strtol (tokens[1], 0, 0);
	      len = sizeof(DataFlash) - start;
	    }
	  if (nt > 2)
	    len = strtol (tokens[2], 0, 0);

	  cprintf("loop %d %d\n", start & 0xff0, (start+len+15)&0xff0);
	  sr = start / 16;
	  er = (start+len+15)/16;
	  for (r=sr; r<er; r++)
	    {
	      cprintf("%w :", r*16);
	      for (i=0; i<16; i++)
		{
		  o = r * 16 + i;
		  if (o >= start && o < start+len)
		    cprintf(" %b", flash_bytes[o]);
		  else
		    tty_puts ("   ");
		  if (i == 7)
		    tty_putc(' ');
		}

	      tty_puts (" : ");
	      for (i=0; i<16; i++)
		{
		  o = r * 16 + i;
		  if (o >= start && o < start+len)
		    {
		      if (isgraph (flash_bytes[o]))
			tty_putc(flash_bytes[o]);
		      else
			tty_putc('.');
		    }
		  else
		    tty_putc (' ');
		  if (i == 7)
		    tty_putc(' ');
		}
	      tty_putc('\n');
	    }
	}

      else if (strcmp (tokens[0], "m") == 0)
	{
	  int start, nstart;

	  if (nt < 2)
	    {
	      cprintf("usage: m address [bytes]");
	      continue;
	    }

	  about_to_modify ();

	  start = strtol (tokens[1], 0, 0);
	  if (nt > 2)
	    {
	      int i;
	      for (i=2; i<nt; i++)
		start = do_m (start, tokens[i], sizeof(DataFlash)-start);
	    }
	  else
	    {
	      while (1)
		{
		  cprintf("%w : ", start);
		  get_line (cmdline, sizeof(cmdline));
		  nstart = do_m (start, cmdline, sizeof(DataFlash)-start);
		  if (start == nstart)
		    break;
		  start = nstart;
		}
	    }
	}

      else if (strcmp (tokens[0], "set") == 0)
	{
	  const VarData *vd;
	  int start, nstart, end;

	  if (nt < 2)
	    {
	      cprintf("usage: set variable bytes...");
	      continue;
	    }

	  vd = str2var(tokens[1], &start);
	  if (!vd)
	    continue;
	  start += vd->offset;
	  end = start + vd->size;

	  about_to_modify ();

	  if (nt > 2)
	    {
	      int i;
	      for (i=2; i<nt; i++)
		start = do_m (start, tokens[i], end-start);
	    }
	  else
	    {
	      while (1)
		{
		  cprintf("%w : ", start);
		  get_line (cmdline, sizeof(cmdline));
		  nstart = do_m (start, cmdline, end-start);
		  if (nstart >= end)
		    break;
		  if (start == nstart)
		    break;
		  start = nstart;
		}
	    }
	}

      else if (strcmp (tokens[0], "copy") == 0)
	{
	  int src, dest;

	  if (nt < 3)
	    {
	      cprintf("usage: copy [source channel] [dest channel|all]\n");
	      continue;
	    }

	  about_to_modify ();

	  src = atoi (tokens[1]);
	  if (strcmp (tokens[2], "all") == 0)
	    {
	      for (dest=0; dest<16; dest++)
		if (dest != src)
		  flash_ptr->power[dest] = flash_ptr->power[src];
	    }
	  else
	    {
	      dest = atoi (tokens[2]);
	      flash_ptr->power[dest] = flash_ptr->power[src];
	    }
	}

      else if (strcmp (tokens[0], "cal") == 0
	       || strcmp (tokens[0], "calib") == 0
	       || strcmp (tokens[0], "calibrate") == 0)
	{
	  int chan;

	  if (nt < 2)
	    {
	      cprintf("usage: calibrate [chan]\n");
	      continue;
	    }

	  about_to_modify ();

	  chan = atoi (tokens[1]);
	  cprintf("cal %d %s\n", chan, tokens[1]);
	  calibrate (chan, flash_ptr->power + chan);
	}

      else if (strcmp (tokens[0], "status") == 0)
	{
	  if (flash_ptr == & flash_temp)
	    tty_puts("There are unwritten changes\n");
	  else
	    tty_puts("Flash data is unchanged\n");
	  if (flash_ptr->header.magic == FLASH_MAGIC)
	    tty_puts("Flash data has been initialized\n");
	  else
	    tty_puts("Flash data is uninitialized\n");
	}

      else if (strcmp (tokens[0], "save") == 0)
	{
	  int c;

	  if (flash_ptr == & flash_data)
	    {
	      tty_puts ("No changes need saving\n");
	      continue;
	    }
	  flash_ptr->header.magic = FLASH_MAGIC;
	  save_flash_data ();
	  flash_ptr = & flash_data;

	  for (c=0; c<16; c++)
	    power_init_channel (c, flash_data.power + c);
	  ether_init_uip ();
	  pm_set_timer ();
	}

      else if (strcmp (tokens[0], "init") == 0)
	{
	  flash_data_init ();
	}

      else if (strcmp (tokens[0], "reset") == 0)
	{
	  tty_flush ();
	  wait_ms (5);
	  prcr.prc1 = 1;
	  pm0.pm03 = 1;
	}

      else
	{
	  tty_puts("Flash commands:\n");
	  tty_puts("\tdiag [1-16|eth]\n");
	  tty_puts("\texit\n");
	  tty_puts("\tshow [variable]\n");
	  tty_puts("\tset [variable] data...\n");
	  tty_puts("\tcopy <src> <dest|all>\n");
	  tty_puts("\tcal[ibrate] <chan>\n");
	  tty_puts("\td [start [len]]\n");
	  tty_puts("\tm start [bytes ...]\n");
	  tty_puts("\tstatus\n");
	  tty_puts("\tsave\n");
	  tty_puts("\treset\n");
	  tty_puts("\thelp\n");
	}
    }
}
